
-- # import Property
-- # import E0
-- # import GHC.Base
-- # import GHC.Base-Models
-- # import CoverPrelude

{------------------------------------------------------------------------
    Models
 ------------------------------------------------------------------------}

-- Expr -------------------------------------------------------------------

data Expr : Set where
  Val : Int -> Expr
  Add : Expr -> Expr -> Expr

injExpr : Expr -> "E0:Expr"
injExpr (Val n)	  = "E0.Val" n
injExpr (Add x y) = "E0.Add" (injExpr x) (injExpr y)

-- Op ---------------------------------------------------------------------

data Op : Set where
  PUSH	: Int -> Op
  ADD	: Op

injOp : Op -> "E0:Op"
injOp (PUSH n)	= "E0.PUSH" n
injOp ADD	= "E0.ADD"

modelOp : Model Op "E0:Op"
modelOp (PUSH n) op = Exists (\n' -> op == "E0.PUSH" n' /\ modelInt n n')
modelOp ADD op	    = op == "E0.ADD"

Code  = List Op

injCode : Code -> "GHC.Base:[]" "E0:Op"
injCode = injList injOp

Stack = List Int
injStack = injList injInt

-- eval -------------------------------------------------------------------

eval : Expr -> Int
eval (Val n)	= n
eval (Add x y)	= eval x + eval y

"eval model" : (e : Expr) -> injInt (eval e) == "E0.eval" (injExpr e)
"eval model" (Val n)	= cover-plugin( include "E0" )
"eval model" (Add x y)	= cover-plugin( include "E0"
				      , include "GHC.Num"
				      , "eval model" x
				      , "eval model" y
				      )

-- comp -------------------------------------------------------------------

comp : Expr -> Code
comp (Val n)	= [PUSH n]
comp (Add x y)	= comp x ++ comp y ++ [ADD]

"comp model" : (e : Expr) -> injCode (comp e) == "E0.comp" (injExpr e)
"comp model" (Val n)	= cover-plugin( include "E0" )
"comp model" (Add x y)	= cover-plugin( include "E0"
				      , "comp model" x
				      , "comp model" y
				      , rem
				      )
  where
    rem : (a, b : Code) -> injCode (a ++ b ++ [ADD]) ==
			      "GHC.Base.++" (injCode a) 
				("GHC.Base.++" (injCode b)
				    ("GHC.Base.:" "E0.ADD" "GHC.Base.[]")
				)
    rem a b = cover-plugin(injList, injOp, modelOp, "++ model" modelOp)

-- exec -------------------------------------------------------------------

injPstack : Partial Stack -> "GHC.Base:[]" Int
injPstack = injPartial injStack

exec : Stack -> Code -> Partial Stack
exec s []		      = val s
exec s (PUSH n :: ops)	      = exec (n :: s) ops
exec (m :: n :: s) (ADD :: ops)  = exec (n + m :: s) ops
exec [n] (ADD :: ops)	      = bot
exec [] (ADD :: ops)	      = bot

"exec model" : (s : Stack) -> (ops : Code) ->
		injPstack (exec s ops) == "E0.exec" (injStack s) (injCode ops)
"exec model" s []		      = cover-plugin( include "E0" )
"exec model" s (PUSH n :: ops)	      = cover-plugin( include "E0"
						    , "exec model" (n :: s) ops
						    )
"exec model" (m :: n :: s) (ADD :: ops)  = cover-plugin( include "E0"
						    , include "GHC.Num"
						    , "exec model" (n + m :: s) ops
						    )
"exec model" [n] (ADD :: ops)	      = cover-plugin(include "E0")
"exec model" [] (ADD :: ops)	      = cover-plugin(include "E0")


{------------------------------------------------------------------------
    Proof
 ------------------------------------------------------------------------}

thm : (e : Expr) -> (s : Stack) -> (ops : Code) ->
       exec s (comp e ++ ops) == exec (eval e :: s) ops
thm (Val n) s ops     = fol-plugin()
thm (Add e1 e2) s ops = fol-plugin( thm e1, thm e2
				  , "++ assoc"
				  , (++)
				  , exec
				  )

cor : (e : Expr) -> (s : Stack) -> (ops : Code) ->
       let e'	= injExpr e
	   s'	= injStack s
	   ops'	= injCode ops
       in
	  "E0.exec" s' ("GHC.Base.++" ("E0.comp" e') ops') ==
	  "E0.exec" ("GHC.Base.:" ("E0.eval" e') s') ops'
cor e s ops = fol-plugin( thm
			, "eval model"
			, "comp model"
			, "exec model"
			, "++ model"
			, injList, injInt
			)

