{-
  Framework.agda

  The core type theory.
-}

-- Relations, predicates
Pred : Set -> Type
Pred X = X -> Set

Rel :  Set -> Type
Rel X = X -> X -> Set

-- The empty set - corresponds to absurdity
data False : Set where

-- "If there was a pattern completeness checker it would see this as a 
-- valid definition" -- Ulf
whenFalse :  (X : Set) -> False -> X

Not : Set -> Set 
Not X = X -> False
  
-- The unit set - corresponds to truth
data True : Set where
    tt : True

-- Binary disjoint union - corresponds to disjunction
data Plus (X,Y : Set) : Set  where
    inl : (x : X) -> Plus X Y
    inr : (y : Y) -> Plus X Y

(\/) : Set -> Set -> Set
P \/ Q = Plus P Q

orI1 (P,Q : Set) : P -> P \/ Q
orI1 p = inl p

orI2 (P,Q : Set) : Q -> P \/ Q
orI2 q = inr q

orE (P,Q,R : Set) : P \/ Q -> (P -> R) -> (Q -> R) -> R
orE (inl p) pr qr = pr p
orE (inr q) pr qr = qr q

-- Dependent pairs - corresponds to existential quantifier
Sigma (X : Set) : (Y : X -> Set) -> Set
Sigma {X} Y = {x : X; Y x}

Exists (X : Set) : (P : Pred X) -> Set
Exists {X} P = Sigma P

(/\) : Set -> Set -> Set
P /\ Q = {P; Q}

(<->) : Set -> Set -> Set
P <-> Q = (P -> Q) /\ (Q -> P)

-- Precedence levels of operators
infixl 4 /\
infixl 3 \/
infix 1 <->

-- The set of boolean values, N_2
data Bool : Set where
    false : Bool
    true : Bool

T : Bool -> Set
T true = True
T false = False

-- Decidable relations
DRel : Set -> Set
DRel X = X -> X -> Bool

dRel (X : Set) : DRel X -> Rel X
dRel D = \x y -> T (D x y)

-- Properties of relations
Reflexive (X : Set) : (R : Rel X) -> Set
Reflexive {X} R = (x : X) -> x `R` x

Symmetrical (X : Set) : (R : Rel X) -> Set
Symmetrical {X} R = (x1,x2 : X) -> (x1 `R` x2 -> x2 `R` x1)

Transitive (X : Set) : (R : Rel X) -> Set
Transitive {X} R = (x1,x2,x3 : X) -> (x1 `R` x2 -> x2 `R` x3 -> x1 `R` x3)
  
Substitutive (X : Set) : (R : Rel X) -> Type
Substitutive {X} R = (P : X -> Set) ->
                   (x1,x2 : X) ->
                       x1 `R` x2 -> P x1 -> P x2

EquivRel (X : Set) : (R : Rel X) -> Set
EquivRel {X} R = { isRefl : Reflexive R;
                   isSymm : Symmetrical R;
                   Transitive R }
