
module TypeChecker.Monad.Context where

import Control.Monad
import Control.Monad.Reader

import Syntax.Internal
import TypeChecker.Monad
import TypeChecker.Monad.Errors
import TypeChecker.Context
import TypeChecker.Errors
import Utils.Map
import Utils.Error

-- Context stuff ----------------------------------------------------------

getContext :: TC Context
getContext = liftM envContext ask

-- | Get the value of a name in the context.
lookupVar :: Name -> TC Value
lookupVar x =
    do	gamma <- getContext
	case lookupMap x gamma of
	    Just (Abstract t)	-> --debug ("lookupVar " ++ x ++ " :: " ++ show t) $
				    return $ VId x t
	    Just (Defined v)	-> return v
	    _	    -> failure $ variableNotBoundError x

-- Note: don't export
updateContext :: (Context -> Context) -> Environment -> Environment
updateContext f env = env { envContext = f $ envContext env }

extendContext :: Name -> ContextEntry -> TC a -> TC a
extendContext x v = local (updateContext $ updateMap x v)

extendContextALot :: Context -> TC a -> TC a
extendContextALot ctx = local (updateContext (`plusMap` ctx))

isBound :: Name -> TC Bool
isBound x = do	gamma <- getContext
		return $ x `isInMap` gamma


