
module TypeChecker.With where

import Control.Monad

import Syntax.Internal
import Syntax.Values
import Syntax.Position

import TypeChecker.Monad
import TypeChecker.Monad.Errors
import TypeChecker.Errors
import TypeChecker.Patterns.Inductive
import TypeChecker.Values
import TypeChecker.UnfoldConstants
import TypeChecker.Universes

import Utils.Monads

withClausePatternMismatch t p =
    sayWhere (getRange t) $
	failure $ customError $ "the pattern " ++ show t ++ " should have the same form as its parent pattern " ++ show p

-- | The first argument is a list of general patterns,
--   the second argument is proper patterns of the same type.
--   The function picks out the parts of the first argument corresponding
--   to the variables bound by the second argument.
restrictPatterns :: [Pattern] -> [Term] -> TC [Term]
restrictPatterns [] ts	       = return ts
restrictPatterns ps []	       = return [] -- failure $ customError $ "too few arguments in with clause"
restrictPatterns (p:ps) (t:ts) = (++) <$> restrict p t <*> restrictPatterns ps ts
    where
	restrict :: Pattern -> Term -> TC [Term]
	restrict (VarP _) t	= return [t]
	restrict (WildP _) t	= return []
	restrict (AbsurdP _) t	= return []
	restrict p@(PairP _ p1 p2) t =
	    do	pv <- patView t
		case pv of
		    PVPair _ t1 t2  -> restrictPatterns [p1,p2] [t1,t2]
		    PVWild r	    -> restrictPatterns [p1,p2] [Meta r dontCare, Meta r dontCare]
		    _		    -> withClausePatternMismatch t p
	restrict p@(ConP _ x hs ps) t =
	    do	pv <- patView t
		case pv of
		    PVCon _ y hs' ps'
			| x == y -> restrictPatterns (hs ++ ps) (hs' ++ ps')
		    PVWild r	 -> restrictPatterns (hs ++ ps)
							(replicate (length (hs ++ ps))
								   (Meta r dontCare)
							)
		    _		 -> withClausePatternMismatch t p

restrictClause :: [Pattern] -> Clause -> TC Clause
restrictClause ps (Clause r x ts w) =
    do	ts' <- expandHidden ts
	Clause r x <$> restrictPatterns ps ts' <*> restrictWith ps w
    where
	restrictWith ps (NoWith e)   = return $ NoWith e
	restrictWith ps (With x es cs) = With x es <$> mapM (restrictClause ps) cs
	restrictWith ps AbsurdClause = return AbsurdClause

-- | @createAuxType Delta As B = Delta -> As -> B@
createAuxType :: Tel -> [Type] -> Type -> Type
createAuxType tel ts t = foldr mkPi (foldr vFun t ts) tel
    where
	mkPi (x,w) t = el (piSort (getSort w) (getSort t)) $ vPi w (mkAbs (nameString x) t)

callAuxFunction :: Name -> Hidden -> Tel -> [Value] -> Value
callAuxFunction x h tel vs = foldl vApp (vDef x h)
			   $ vars tel ++ vs
    where
	vars tel = map vBId $ reverse $ [0..length tel - 1]

